import Type from "./type";
import Action from "../action";
import Vue from "vue";

let loading = false;
const dirtyPrompt = "Changes you made have not been saved. Are you sure?";

class HttpRequest {
  constructor() {
    this.controller = new AbortController();
    this.canceled = false;
  }

  cancel() {
    this.canceled = true;
    this.controller.abort();
  }
}

export default class {
  static _extractFormData(properties, needCsrfToken) {
    // if (properties["formDataEncoded"]) {
    //   return properties["formData"];
    // }

    const data = properties.formData;
    if (data instanceof FormData) {
      return data;
    }

    const formData = new FormData();
    var params = data || {};
    for (const key in params) {
      formData.append(key, params[key]);
    }

    if (needCsrfToken && !Type.isString(params["authenticity_token"])) {
      formData.append("authenticity_token", Utils.dom.getCsrf());
    }
    return formData;
  }

  static load(properties, component) {
    const url = new URL(properties["url"]);
    const domainMatched = window.location.hostname == url.hostname;

    // If this is an external domain, we rely on `onUnload()` instead.
    if (domainMatched && !this.proceedEvenWhenDirty()) {
      return;
    }

    if (Utils.settings.reactive && domainMatched) {
      const currentUrl = window.location.href;
      const htmlUrl = Utils.url.htmlUrl(properties["url"]);

      Utils.http.execute(properties, "GET", component, (data, response) => {
        if (htmlUrl !== currentUrl) {
          const redirectUrl = Utils.url.htmlUrl(response.url);
          Utils.history.pushPage(data, redirectUrl);
        }

        this.forceComponentUpdate(() => {
          Utils.history.resetScroll();
          window.vueApp.page = data;
          Action.execute(properties["onOpen"], component);
        });
      });
    } else {
      window.location = Utils.url.htmlUrl(properties["url"]);
    }
    // bus.$emit("glibWebHttpLoad", new URL(Utils.url.htmlUrl(properties["url"])));
  }

  static analyticsHeaders(component) {
    const pageAnalytics =
      window.vueApp.temp.analytics || window.vueApp.page.analytics || {};
    const featureAnalytics = pageAnalytics.feature;

    const headers = {};
    if (featureAnalytics && component.spec) {
      const componentAnalytics = component.spec.analytics || {};
      if (componentAnalytics.disabled) {
        return headers;
      }

      Utils.type.ifString(
        componentAnalytics.group || featureAnalytics.group,
        value => (headers["GApp-Analytics-Referer-Group"] = value)
      );

      Utils.type.ifString(
        componentAnalytics.action || featureAnalytics.action,
        value => (headers["GApp-Analytics-Referer-Action"] = value)
      );

      Utils.type.ifString(
        componentAnalytics.segment || featureAnalytics.segment,
        value => (headers["GApp-Analytics-Referer-Segment"] = value)
      );

      Utils.type.ifString(
        componentAnalytics.placement,
        value => (headers["GApp-Analytics-Referer-Placement"] = value)
      );
    }
    return headers;
  }

  static reload(properties, component) {
    const currentUrl = window.location.href;
    const data = {
      url: properties.url || currentUrl
    };

    Utils.http.execute(data, "GET", component, (page, response) => {
      Utils.http.forceComponentUpdate(() => {
        window.vueApp.page = page;
        const redirectUrl = Utils.url.htmlUrl(response.url);
        Utils.history.updatePage(redirectUrl);

        GLib.action.execute(properties["onReload"], component);
      });
    });
  }

  static execute(properties, methodName, component, jsonHandler) {
    this.startIndicator(component);

    // `fetch()` only supports uppercase
    const method = methodName.toUpperCase();
    let url = Utils.url.jsonUrl(properties["url"]);
    let body = this._extractFormData(properties, method !== "GET");
    if (method === "GET") {
      url = Utils.url.appendParams(url, body);
      body = null;
    }
    console.log(`${method} ${url}`);

    const request = new HttpRequest();
    let response = null;
    const vm = this;

    const analyticsHeaders = this.analyticsHeaders(component);

    fetch(url, {
      method: method,
      body: body,
      headers: Object.assign({ Accept: "text/html" }, analyticsHeaders),
      // Make sure to pass cookies for same origin URLs.
      // Needed for some browsers, e.g. Edge and Android's native.
      credentials: "same-origin",
      signal: request.controller.signal
    })
      .then(res => {
        vm.stopIndicator(component);

        if (res.status >= 500) {
          throw "Server error";
        } else if (res.status >= 400) {
          throw "Not accessible";
        } else {
          response = res;
          return res.json();
        }
      })
      .then(data => {
        console.debug("Success", data);
        jsonHandler(data, response);
      })
      .catch(error => {
        vm.stopIndicator(component);
        if (!request.canceled) {
          console.error("Error:", error);
          Utils.launch.snackbar.error(error, component);
        } else {
          console.info("Canceled");
        }
      });

    return request;
  }

  static startIndicator(component) {
    this._showIndicator();
    component.$data._isBusy = true;
  }

  static stopIndicator(component) {
    this._hideIndicator();
    component.$data._isBusy = false;
  }

  static _showIndicator() {
    loading = true;
    setTimeout(() => {
      if (loading) {
        loading = false;
        window.vueApp.indicator = true;
      }
    }, 200);
  }

  static _hideIndicator() {
    loading = false;
    window.vueApp.indicator = false;
  }

  // See generic.js
  static forceComponentUpdate(handler) {
    // GLib.component.clearRegistry();

    const dirtyContext = this.dirtyContext();

    window.vueApp.isStale = true;

    // Queue the execution so the first isStale has time to resets state before handler gets executed
    Vue.nextTick(() => {
      dirtyContext.isFormSubmitted = false;
      dirtyContext.isFormDirty = false;
      handler();
      window.vueApp.isStale = false;
    });
  }

  static promptIfDirtyOnUnload() {
    window.onbeforeunload = () => {
      return this.dirtyContext().isFormDirty ? dirtyPrompt : null;
    };
  }

  static dirtyContext() {
    return Utils.launch.dialog.topDialog() || window.vueApp;
  }

  static clearDirtyState() {
    this.dirtyContext().isFormDirty = false;
  }

  static notifyFormSubmitted() {
    this.dirtyContext().isFormSubmitted = true;
  }

  // `context` can be either window or dialog.
  static proceedEvenWhenDirty() {
    const dirtyContext = this.dirtyContext();
    // Don't prompt if this is a result of form submission
    if (
      dirtyContext.isFormDirty &&
      !dirtyContext.isFormSubmitted &&
      !confirm(dirtyPrompt)
    ) {
      return false;
    }
    return true;
  }
}
